/*******************************************************************************
 * eppmanual.c
 * Functions for software-mode EPP communication using a parallel port
 *
 * July 16, 2005
 *
 * Michael Siegenthaler
 * www.msigi.net
 *
 ******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/io.h>

#include "eppmanual.h"

int epp_port_has_been_opened = 0;

/*******************************************************************************
 * Write to a memory location on the device
 * Parameters:  address to write to, data to write
 * Returns 0 on success, 1 on failure
 ******************************************************************************/
int write_dev(int addr, int data) {

  int retry_count;
  unsigned char value;

  if (!epp_port_has_been_opened) {
    if (open_port()) return 1;
  }

  // Address
  retry_count = 0;
  while (1) {

    // Write address
    write_port((unsigned char)addr, 1);

    // Read address back
    read_port(&value, 1);

    // Check
    if (value==(unsigned char)addr) break;
    retry_count++;
    if (retry_count>=MAX_EPP_RETRIES) return 1;

  }

  // Data
  retry_count = 0;
  while (1) {

    // Write data
    write_port((unsigned char)data, 0);

    // Read data back
    read_port(&value, 0);

    // Check
    if (value==(unsigned char)data) break;
    retry_count++;
    if (retry_count>=MAX_EPP_RETRIES) return 1;

  }

  return 0;

}

/*******************************************************************************
 * Read from a memory location on the device
 * Parameters:  address to read, pointer to location for storing data
 * Returns 0 on success, 1 on failure
 ******************************************************************************/
int read_dev(int addr, int *data) {

  int retry_count;
  unsigned char data1, data2, value;

  if (!epp_port_has_been_opened) {
    if (open_port()) return 1;
  }

  // Address
  retry_count = 0;
  while (1) {

    // Write address
    write_port((unsigned char)addr, 1);

    // Read address back
    read_port(&value, 1);

    // Check
    if (value==(unsigned char)addr) break;
    retry_count++;
    if (retry_count>=MAX_EPP_RETRIES) return 1;

  }

  // Data
  retry_count = 0;
  while (1) {

    // Read data twice
    read_port(&data1, 0);
    read_port(&data2, 0);

    // Check
    if (data1==data2) break;
    retry_count++;
    if (retry_count>=MAX_EPP_RETRIES) return 1;

  }

  // Return result
  *data = (int) data1; // Convert to integer
  *data = *data & 255;  // Drop leading 1's in case it was sign-extended

  return 0;

}

/*******************************************************************************
 * Open port
 * Returns 0 on success, 1 on failure
 ******************************************************************************/
int open_port(void) {

  // Make sure we have permission to open the port
  if (ioperm(BASE,1,1))
    return 1;
  if (ioperm(BASE+1,1,1))
    return 1;
  if (ioperm(BASE+2,1,1))
    return 1;

  epp_port_has_been_opened = 1;
  return 0;

}

/*******************************************************************************
 * Write an address or data
 * Parameters: value to write, boolean indicating whether it is an address
 * Returns 0 on success, currently does not detect a failure
 ******************************************************************************/
int write_port(unsigned char value, int is_address) {

  int c;
  unsigned char control;

  // Pull nReset line high, nWrite low
  control = 0x05; // 0000 0101 -> xxxx 1110
  outb((unsigned char)control, BASE+2);

  // Put data on output pins
  outb((unsigned char)value, BASE);

  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // At this point we should make sure that the device is not asserting the Wait
  // line, but in the interest of laziness, we ignore it. This is not a problem,
  // because the device is clocked faster than the parallel port can provide data.

  // Pull strobe low
  if (is_address) {
    control = 0x0d; // 0000 1101 -> xxxx 0110
  } else {
    control = 0x07; // 0000 0111 -> xxxx 1100
  }
  outb((unsigned char)control, BASE+2);
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // At this point we should wait for the device to assert the Wait line, but
  // again, this is ignored at least for now.

  // Pull strobe high
  control = 0x05; // 0000 0101 -> xxxx 1110
  outb((unsigned char)control, BASE+2);
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // Go back into reset state
  set_reset();
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  return 0;

}

/*******************************************************************************
 * Read an address or data
 * Parameters: pointer to location to store the value read, boolean indicating
 *   whether it is an address
 * Returns 0 on success, currently does not detect a failure
 ******************************************************************************/
int read_port(unsigned char* value, int is_address) {

  int c;
  unsigned char readval, control;

  // put port in bidir, non-reset, read mode
  control = 0x24; // 0010 0100 -> xxxx 1111
  outb((unsigned char)control, BASE+2);
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // pull strobe low
  if (is_address) {
    control = 0x2c; // 0010 1100 -> xxxx 0111
  } else {
    control = 0x26; // 0010 0110 -> xxxx 1101
  }
  outb((unsigned char)control, BASE+2);
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // At this point, we should wait until Wait is asserted, but this can
  // be ignored because the device is fast

  // read data
  readval = inb(BASE);
  *value = readval;

  //printf("got this: %x\n", readval);

  // pull strobe high again
  control = 0x24; // 0010 0100 -> xxxx 1111
  outb((unsigned char)control, BASE+2);
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  // Go back into reset state
  set_reset();
  for (c=0;c<BUSYLOOPCOUNT;c++);;

  return 0;

}

/*******************************************************************************
 * Put port into reset state
 ******************************************************************************/
void set_reset() {
  char control;

  // Pull nReset line low, others high
  control = 0x00; // 0000 0000 -> xxxx 1011
  outb((unsigned char)control, BASE+2);

}
